"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createTaskRunner = void 0;
const cron_parser_1 = require("cron-parser");
const promises_1 = require("fs/promises");
const path_1 = require("path");
const tiny_glob_1 = __importDefault(require("tiny-glob"));
const typescript_1 = require("typescript");
async function loadTasks(folder, transpiledFolder, pattern) {
    if (!(0, path_1.isAbsolute)(folder))
        throw new Error('folder must be an absolute path');
    const files = await (0, tiny_glob_1.default)(pattern, { cwd: folder });
    const tasks = [];
    await Promise.all(files.map(async (file) => {
        const modulePath = (0, path_1.join)(folder, file).replace(/\\/g, '/');
        const isTypescript = modulePath.endsWith('.ts');
        let module;
        if (isTypescript) {
            // TODO: add caching with md5 of source.
            const content = await (0, promises_1.readFile)(modulePath, 'utf-8');
            const transpiledPath = (0, path_1.join)(transpiledFolder, file.replace('.ts', '-transpiled.js')).replace(/\\/g, '/');
            const transpiledCode = (0, typescript_1.transpileModule)(content, {}).outputText;
            await (0, promises_1.mkdir)((0, path_1.dirname)(transpiledPath), { recursive: true });
            await (0, promises_1.writeFile)(transpiledPath, transpiledCode);
            module = require(transpiledPath);
        }
        else {
            module = require(modulePath);
        }
        const isValid = 'default' in module && typeof module.default === 'object';
        if (!isValid) {
            console.error('Error loading task file:', file, 'it must contain single default export:');
            console.error('     ```````````````````````````````````````````````````````````');
            console.error("     ` const task = taskBuilder.daily().at('11:30').build(); `");
            console.error('     ` export default task                                     `');
            console.error('     ```````````````````````````````````````````````````````````');
            return;
        }
        tasks.push({ ...module.default, path: file });
    }));
    console.info('Loaded', tasks.length, 'tasks from folder', folder);
    // TODO: use tiny-glob
    return tasks;
}
function shouldRunTask(task, tickTime) {
    if (task.runsAt) {
        // [tick1] ... [tick2] <->. [runsAt] ....
        if (task._executed)
            return;
        if (Date.now() > task.runsAt.getTime() - tickTime) {
            return true;
        }
    }
    else {
        const cron = (0, cron_parser_1.parseExpression)(task.cron);
        const nextSuitable = cron.next();
        if (Date.now() > nextSuitable.toDate().getTime() - tickTime) {
            return true;
        }
    }
    return false;
}
function taskToString(task) {
    if (task.name)
        return `${task.name} (${task.cron || task.runsAt?.toLocaleString()})`;
    if (task.path)
        return `${task.path} (${task.cron || task.runsAt?.toLocaleString()})`;
    return `Unknown task (${task.cron || task.runsAt?.toLocaleString()})`;
}
async function createTaskRunner(options) {
    const tasks = options.tasks || (await loadTasks(options.folder, options.transpiledFolder, options.pattern));
    const tickTime = options.tickTime || 1000;
    const taskNames = new Set();
    for (const task of tasks) {
        if (taskNames.has(task.name))
            throw new Error(`Task "${task.name}" is defined twice, this is not allowed.`);
        taskNames.add(task.name);
    }
    const tick = () => {
        tasks.forEach(task => {
            if (!shouldRunTask(task, tickTime))
                return;
            if (task.runsAt) {
                task._executed = true;
            }
            try {
                task.execute()
                    .then(() => {
                    options.onTaskComplete?.(task);
                })
                    .catch(err => {
                    console.error('\x1b[31m[Error] at task', taskToString(task), '\x1b[0m');
                    console.error(err);
                });
            }
            catch (err) {
                console.error('Error running task', taskToString(task), err);
            }
        });
    };
    const tid = setInterval(tick, tickTime);
    const stop = () => clearInterval(tid);
    const appendTasks = (tasks) => {
        let appendedAmount = 0;
        for (const newTask of tasks) {
            if (taskNames.has(newTask.name)) {
                console.info('Skipping', newTask.name, 'as it already exists');
                continue;
            }
            tasks.push(newTask);
            appendedAmount++;
        }
        console.info('Appended', appendedAmount, 'new tasks!');
    };
    return { tid, stop, appendTasks };
}
exports.createTaskRunner = createTaskRunner;
